"""Forward modelling sub-package — 1D and 2D resistivity simulation."""
from georesistpy.forward.forward1d import forward_1d  # noqa: F401
from georesistpy.forward.forward2d import forward_2d  # noqa: F401
