"""Utility sub-package — array configs, DOI, topography."""
from georesistpy.utils.arrays import (  # noqa: F401
    generate_array,
    ARRAY_TYPES,
)
from georesistpy.utils.doi import compute_doi  # noqa: F401
